# RA, 2021-03-17

class Peek:
    def __init__(self, reporter=None):
        self.reporter = reporter

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def __call__(self, x=None):
        if self.reporter is not None:
            self.reporter(x)
        return x



# RA, 2020-10-09
# RA, 2020-12-06
# RA, 2020-10-19


import io
import os
import pathlib
import typing
import contextlib


def relpath(path):
    return os.path.relpath(str(path), os.getcwd())


def assert_exists(file):
    if not pathlib.Path(file).exists():
        raise FileNotFoundError(F"Oh, file `{file}`, where art thou?!")
    else:
        return file


@contextlib.contextmanager
def seek_then_rewind(fd: io.IOBase, seek=None) -> io.IOBase:
    pos = fd.tell()
    if seek is not None:
        fd.seek(seek)
    try:
        yield fd
    finally:
        fd.seek(pos)


# Adapted from
# https://stackoverflow.com/a/4213255
def md5(fd: io.TextIOBase) -> str:
    import hashlib
    with seek_then_rewind(fd):
        md5 = hashlib.md5()
        for chunk in iter(lambda: fd.read(128 * md5.block_size).encode(), b''):
            md5.update(chunk)
    return md5.hexdigest()


def mkdir(path: pathlib.Path, parents=True, exists_ok=True):
    assert not path.is_file(), F"Cannot create folder (it is a file: {path})."
    path.mkdir(parents=parents, exist_ok=exists_ok)
    return path



# RA, 2020-10-09
# RA, 2020-12-06

def from_iterable(X, type=None):
    from itertools import chain
    if type is None:
        return chain.from_iterable(X)
    else:
        return type(chain.from_iterable(X))


def first(X):
    """
    Returns the first element of an iterable X.
    """
    return next(iter(X))


def at_most_n(X, n=-1):
    """
    Yields at most n elements from iterable X.
    All elements if n equals -1.
    No elements if n is None.
    """
    if n is None:
        pass
    elif (n == -1):
        yield from X
    else:
        for (x, __) in zip(X, range(n)):
            yield x


def unlist1(L):
    """
    Check that L has only one element at return it.
    """
    L = list(L)
    if not (len(L) == 1):
        raise ValueError(F"Expected an iterable of length 1, got {len(L)}.")
    return L[0]


def pretty(x):
    """
    Return a formatted JSON string of `x`.
    """
    import json
    return json.dumps(x, indent=2)


class minidict:
    """
    A slim read-only dictionary.
    """

    def __init__(self, data: dict):
        self._data = dict(data)

    def __repr__(self):
        return repr(self._data)

    def __getitem__(self, item):
        return self._data[item]

    def __len__(self):
        return len(self._data)

    def keys(self):
        return self._data.keys()

    def items(self):
        return zip(self._data.keys(), map(self.__getitem__, self._data.keys()))

    def __contains__(self, item):
        return (item in self._data)
